﻿var jsCalc = (function() {
	var out = null; //must be an object with methods "empty" and "append"
	var jsEval = eval; //in case eval is overwritten
	var lines = 4; //visible lines count in output window

	function print() {
		if (!out) {
			alert('Output not connected!');
			return;
		}
	
		var fn = print.fix ? Number.fixedFormat(print.digits) : Number.floatFormat(print.digits);
		for (var i = 0; i < arguments.length; i++)
			out.append(Number.objToStr(arguments[i], fn));
		
		if (print.term)
			out.append(print.term);
	}
	
	function exec(program) {
		if (!out) {
			alert('Output not connected!');
			return;
		}
	
		out.empty();
		var result = jsEval(program);
		if (typeof result !== 'undefined')
			print(result);
			
		if (Number.isNumber(print.lines)) {
			if (print.lines < 4)
				print.lines = 4;
			else if (print.lines > 100)
				print.lines = 100;
			
			if (lines != print.lines) {
				lines = print.lines;
				out.css('height', (1.2 * lines + 1) + 'em')
			}
		}
	}

	function connect(obj) {
		out = obj;
	}

	return { exec: exec, print: print, connect: connect };
})();

eval = undefined; //free to use as variable name

/* CONSTANTS */
function loadConst(withUnits) {
	alpha = 0.0072973525664;
	pi = Math.PI;
	π = pi;
	e = Math.E;
	deg = pi/180;
	
	if (withUnits) {
		c = Qty(299792458, 'm s^-1');
		h = Qty(6.626070040E-34, 'J s');
		hbar = Qty(1.0545718e-34, 'J s');
		eps_0 = Qty(8.85418781762038982e-12, 'F m^-1');
		mu_0 = Qty(12.566370614359173e-7, 'N A^-2');
		G = Qty(6.67408e-11, 'N m^2 kg^-2');
		q_e = Qty(1.6021766208e-19, 'C');
		m_e = Qty(9.10938356e-31, 'kg');
		k_B = Qty(1.38064852e-23, 'J K^-1');
		R = Qty(8.314472, 'J mol^-1 K^-1');
		g = Qty(9.80665, 'm s^-2');
		sigma = Qty(5.670367e-8, 'W m^-2');
		b = Qty(0.0028977729, 'm K');
		N_A = Qty(6.022140857e23, 'mol^-1');
		M_e = Qty(5.9722e24, 'kg');
		M_s = Qty(1.98855e30, 'kg');
	} else {
		c = 299792458;
		h = 6.626070040E-34;
		hbar = 1.0545718e-34;
		eps_0 = 8.85418781762038982e-12;
		mu_0 = 12.566370614359173e-7;
		G = 6.67408e-11;
		q_e = 1.6021766208e-19;
		m_e = 9.10938356e-31;
		m_p = 1.672621898e-27;
		m_n = 1.674927471e-27;
		k_B = 1.38064852e-23;
		R = 8.314472;
		g = 9.80665;
		sigma = 5.670367e-8;
		b = 0.0028977729;
		N_A = 6.022140857e23;
		at = 101325;
		eV = 1.6021766208e-19;
		lb = 0.45359237;
		inch = 0.0254;
		foot = 0.3048;
		psi = 6894.757;
		mile = 1609.344;
		mph = 0.44704;
		kph = 0.277777777777777778;
		M_e = 5.9722e24;
		M_s = 1.98855e30;	
	}
	
	ε_0 = eps_0;
	μ_0 = mu_0;
	σ = sigma;
	ħ = hbar;
}

loadConst(false);

/* MATH FUNCTIONS */
var abs = Math.abs;
var acos = Math.acos;
var asin = Math.asin;
var sin = Math.sin;
var cos = Math.cos;
var tan = Math.tan;
var sinh = Math.sinh;
var cosh = Math.cosh;
var tanh = Math.tanh;
var sqrt = Math.sqrt;
var exp = Math.exp;
var ln = Math.log;
var pow = Math.pow;
var ceil = Math.ceil;
var floor = Math.floor;
var round = Math.round;
var range = Array.range;
var linspace = Array.linspace;
var sum = Math.sum;
var mean = Math.mean;
var stdev = Math.stdev;


function log() {
	if (arguments.length == 1)
		return Math.log(arguments[0]);
	
	if (arguments.length == 2)
		return Math.log(arguments[0]) / Math.log(arguments[1]);
		
	throw "Function log() takes 1 or 2 arguments";
}

function atan() {
	if (arguments.length == 1)
		return Math.atan(arguments[0]);
		
	if (arguments.length == 2)
		return Math.atan2(arguments[1], arguments[0]);
	
	throw "Function atan() takes 1 or 2 arguments";
}

function rnd() {
	if (arguments.length == 0)
		return Math.random();
	
	if (arguments.length == 1)
		return Math.random() * arguments[0];
		
	var min = arguments[0];
	var max = arguments[1];
	return min + Math.random() * (max - min);
}

var random = rnd;

function minval() {
	var count = 0;
	var min = Number.POSITIVE_INFINITY;
	
	function parse(array) {
		for (var i = 0; i < array.length; i++) {
			var val = array[i];
			if (Number.isNumber(val)) {
				count++;
				if (val < min)
					min = val;
			}
			else if (Array.isArray(val))
				parse(val);
		}
	}

	parse(arguments);
	if (count == 0)
		throw "minval: at least one numerical argument required";
	
	return min;
};

var min = minval;


function maxval() {
	var count = 0;
	var max = Number.NEGATIVE_INFINITY;
	
	function parse(array) {
		for (var i = 0; i < array.length; i++) {
			var val = array[i];
			if (Number.isNumber(val)) {
				count++;
				if (val > max)
					max = val;
			}
			else if (Array.isArray(val))
				parse(val);
		}
	}

	parse(arguments);
	if (count == 0)
		throw "maxval: at least one numerical argument required";
	
	return max;
};

var max = maxval;

	
Random = {
	uniform: Math.random,
	
	unif: Math.random,
	
	gauss: function() {
		var x1;
		var x2;
		var w;
		while (true) {
			x1 = 2 * Math.random() - 1;
			x2 = 2 * Math.random() - 1;
			w = x1 * x1 + x2 * x2;
			if (w < 1)
				break;
		};
		
		return x1 * Math.sqrt( (-2 * Math.log(w) ) / w );
	}
}

Special = {
	erf: function(x) {
		var s = Math.sign(x);
		x = Math.abs(x);
		var t = 1 / (1 + 0.3275911 * x);
    	return s * ( 1 - (((((1.061405429 * t - 1.453152027) * t) + 1.421413741) * t - 0.284496736) * t + 0.254829592) * t * Math.exp(-x * x) );
	},
	
	cheb: Cheb.base,
	
	gamma: (function(){
		var g = 9;
		var C = [0.99999999999980993, 676.5203681218851, -1259.1392167224028, 771.32342877765313, -176.61502916214059, 12.507343278686905, -0.13857109526572012, 9.9843695780195716e-6, 1.5056327351493116e-7];
		
		function gamma(x) {
			if (x < 0.5)
				return Math.PI / (Math.sin(Math.PI * x) * gamma(1 - x));

			x -= 1;
			var a = C[0];
			for (var i = 1; i < g; i++)
				a += C[i] / (x + i);
			
			var t = x + g - 1.5;
			return a * Math.sqrt(2 * Math.PI) * Math.pow(t, x + 0.5) * Math.exp(-t);
		};
		
		return gamma;
	})()
};

var Fit = {
	line: function(x, y) {
		if (!Array.isArray(x) || !Array.isArray(y))
			throw "Fit.line: invalid arguments (arrays expected)";
			
		var N = Math.min(x.length, y.length);
		if (N < 2)
			throw "Fit.line: not enough data points";
		
		var Σx = 0; var Σy = 0; var Σxy = 0; var Σxx = 0; var Σyy = 0;
		for (var i = 0; i < N; i++) {
			Σx += x[i];
			Σy += y[i];
			Σxy += x[i] * y[i];
			Σxx += x[i] * x[i];
			Σyy += y[i] * y[i];
		}
		
		/* hereafter the symbols indicate means rather than sums */
		Σx /= N; Σy /= N; Σxy /= N; Σxx /= N; Σyy /= N;
		
		var vxx = Σxx - Σx * Σx;
		var vyy = Σyy - Σy * Σy;
		var vxy = Σxy - Σx * Σy;
		var slope = vxy / vxx;
		var offset = Σy - slope * Σx;
		
		/*for standard error*/
		var s = Math.sqrt((vyy - slope * vxy) / (N - 2));
		var eslope = s / Math.sqrt(vxx);
		var eoffset = s * Math.sqrt(1 + Σx * Σx / vxx);
		
		var line = Polynom([offset, slope]);
		line.slope = slope;
		line.offset = offset;
		line.eslope = eslope;
		line.eoffset = eoffset;
		return line;
	},
	
	lincomb: function(x,y,F) {
		if (!Vector.isValidArray(x) || !Vector.isValidArray(y) || x.length !== y.length)
			throw "Fit.lincomb: arguments must be non-empty arrays of equal length";
		
		var Q;
		var R;
		if (Array.isArray(F) && F.every(Function.isFunction)) {
			var data = [];
			F.forEach(function(f) {
				data.push( x.map(f) );
			});
			
			Q = Matrix._create(data);
			R = Q.trans();
		}
		else if (Function.isFunction(F)) {
			R = Matrix._create(x.map(F));
			Q = R.trans();
		} else
			throw "Fit.lincomb: F must be an array of functions";
		
		var A = Q.mul(R);
		var b = Q.mul( Vector._create(y) );
		return Matrix.lsolve(A,b);
	},
	
	spline: function(x,y) {
		if (!Vector.isValidArray(x) || !Vector.isValidArray(y) || x.length !== y.length)
			throw "Fit.spline: arguments must be non-empty arrays of equal length";
		
		var n = x.length - 1; //index of the last data point (not length)
		var dx = [];
		var dy = [];
		for (var i = 0; i < n; i++) {
			var d = x[i+1] - x[i];
			if (d <= 0)
				throw "Fit.spline: x-values must be in increasing order";
			dx.push( d );
			dy.push( y[i+1] - y[i] );
		}
		
		var A = [];
		var v = [];
		var row;
		for (var i = 1; i < n; i++) {
			row = Array.const(n+1);
			row[i-1] = 1/dx[i-1];
			row[i] = 2/dx[i-1] + 2/dx[i];
			row[i+1] = 1/dx[i];
			A.push(row);
			v.push( 3 * dy[i-1] / (dx[i-1] * dx[i-1]) + 3 * dy[i] / (dx[i] * dx[i]) );
		}
		
		row = Array.const(n+1);
		row[0] = 2/dx[0];
		row[1] = 1/dx[0];
		A.push(row);
		v.push( 3 * dy[0] / (dx[0] * dx[0]) );
		
		row = Array.const(n+1);
		row[n-1] = 1/dx[n-1];
		row[n] = 2/dx[n-1];
		A.push(row);
		v.push( 3 * dy[n-1] / (dx[n-1] * dx[n-1]) );
		var k = Matrix.lsolve(A,v);
		
		var a = [];
		var b = [];
		for (var i = 0; i < n; i++) {
			a.push( k[i] * dx[i] - dy[i] );
			b.push( -k[i+1] * dx[i] + dy[i] );
		}
		
		return function(p) {
			if (p <= x[0])
				return y[0] + k[0] * (p - x[0]);
			
			if (p >= x[n])
				return y[n] + k[n] * (p - x[n]);
			
			var i = n-1;
			while (x[i] > p)
				i--;
		
			var t = (p - x[i]) / dx[i];
			return (1-t)*y[i] + t*y[i+1] + t*(1-t)*(a[i]*(1-t)+b[i]*t);
		};
	},
	
	poly: function(x, y, degree) {
		var F = Polynom.baseSet(degree);
		var coeff = Fit.lincomb(x, y, F);
		var pol = Polynom(coeff);
		pol.coeff = coeff;		
		return pol;		
	},
	
	cheb: function(x, y, degree) {
		var F = Cheb.baseSet(degree);
		var coeff = Matrix.linfit(x, y, F);
		var pol = Cheb(coeff);
		pol.coeff = coeff;		
		pol.toStr = function(fn) { return 'y = ' + Cheb.format(coeff, fn, 'x'); };
		return pol;		
	}
}

var Solve = {
	quad: function(a, b, c) {
		if (typeof a === 'object')
			a = a.valueOf();
		
		if (typeof b === 'object')
			b = b.valueOf();
		
		if (typeof c === 'object')
			c = c.valueOf();
		
		if (!Number.isFinite(a) || !Number.isFinite(b) || !Number.isFinite(c))
			throw "Solve.quad: invalid arguments (finite numbers expected)";
		
		if (a === 0) {
			if (b === 0) {
				var sol = [];
				sol.toStr = function(fn) {
					return c === 0 ? "equation is satisfied by every value of x" : "equation has no solutions";
				};
				return sol;
			}
		
			x = -c / b;
			var sol = [x];
			sol.toStr = function(fn) {
				return "x = " + fn(x);
			};
			return sol;
		}
		
		var D = b*b - 4*a*c;
		if (D === 0) {
			var x = -b/(2*a);
			var sol = [x];
			sol.toStr = function(fn) {
				return "x = " + fn(x);
			};
			return sol;
		}
		
		if (D > 0) {
			D = Math.sqrt(D);
			var x1 = (-b+D)/(2*a);
			var x2 = (-b-D)/(2*a);
			var sol = [x1, x2];
			sol.toStr = function(fn) {
				return "x1 = " + fn(x1) + ", x2 = " + fn(x2);
			};
			return sol;
		}
		
		D = Math.sqrt(-D);
		re = -b/(2*a);
		im = D/(2*a);
		var x1 = Complex(re, im);
		var x2 = Complex(re, -im);
		var sol = [x1, x2];
		sol.toStr = function(fn) {
			return "x1 = " + x1.toStr(fn) + ", x2 = " + x2.toStr(fn);
		};
		return sol;
	},
	
	linear: Matrix.lsolve,
	
	root: function(f, a, b, eps) {
		if (!Function.isFunction(f))
			throw "Solve.root: f must be a function";
		
		if (typeof a === 'object')
			a = a.valueOf();
		
		if (typeof b === 'object')
			b = b.valueOf();
		
		if(!Number.isFinite(a) || !Number.isFinite(b))
			throw "Solve.root: a and b must be finite numbers";
		
		var f_a = f(a);
		var f_b = f(b);
		
		if (f_a * f_b > 0)
			throw "Solve.root: invalid range, f(a) and f(b) have the same sign";
		
		if (typeof eps === 'undefined')
			eps = 1e-4;
		else {
			if (typeof eps === 'object')
				eps = eps.valueOf();

			if (!Number.isFinite(eps))
				throw "Solve.root: eps must be a finite number";
		}
		
		n = 0;
		var c;
		var f_c;
		while (abs(a - b) > eps) {
			//c = (a*f_b - b*f_a) / (f_b - f_a);
			c = (a + b) / 2;
			f_c = f(c);
			
			if (f_a * f_c < 0) {
				b = c;
				f_b = f_c;
			} else {
				a = c;
				f_a = f_c;
			}
			n++;
		}
		
		var sol = new Number(c);
		sol.error = Math.abs(b-a);
		sol.itercount = n;
		return sol;
	},
	
	opt: function(f, x, d, n, s) {
		if (!Function.isFunction(f))
			throw "Solve.opt: f must be a function";
		
		if (Number.isNumber(x) || Number.isNumber(d)) {
			x = [x];
			d = [d];
		}
		else if (!Array.isArray(x) || !Array.isArray(d) || x.length !== d.length || x.length === 0)
			throw "Solve.opt: x and d must be numbers or non-empty arrays of equal length";
		
		if (!Number.isFinite(n) || n <= 0)
			throw "Solve.opt: n must be a positive number";
		
		if (!Number.isFinite(s))
			throw "Solve.opt: s must be a finite number";
		
		var M = s * f.apply(null, x);
		var m;
		for (var i = 0; i < n; i++) {
			for (var k = 0; k < x.length; k++) {
				var old = x[k];
				x[k] = old + d[k];
				m = s * f.apply(null, x);
				if (m < M)
					M = m;
				else {
					x[k] = old - d[k];
					m = s * f.apply(null, x);
					if (m < M)
						M = m;
					else {
						x[k] = old;
						d[k] *= 0.5;
					}
				}
			}
		}
		
		return M;
	},
	
	min: function(f, x, d, n) { return this.opt(f, x, d, n, 1); },
	max: function(f, x, d, n) { return this.opt(f, x, d, n, -1); },
	
	diff: function(f, y0, tmax, dt) {
		if (!Function.isFunction(f))
			throw "Solve.diff: f must be a function";
			
		var deriv;
		
		if (Array.isArray(y0)) {
			if (!y0.every(Number.isFinite))
				throw "Solve.diff: y0 must be an array of finite numbers";
			
			deriv = f;
		} else {
			if (typeof y0 === 'object')
				y0 = y0.valueOf();
			
			if (!Number.isFinite(y0))
				throw "Solve.diff: y0 must be a finite number";
			
			y0 = [y0];
			deriv = function(t,y) { return [ f(t,y) ]; };
		}
		
		if (typeof tmax === 'object')
			tmax = tmax.valueOf();
		
		if (!Number.isFinite(tmax) || tmax <= 0)
			throw "Solve.diff: tmax must be a positive finite number";
		
		if (typeof dt === 'object')
			dt = dt.valueOf();
		
		var dt_2 = 0.5 * dt;
		
		if (!Number.isFinite(dt) || dt <= 0)
			throw "Solve.diff: the step must be a positive finite number";
		
		var t = 0;
		var y = y0;
		var sol = [ [t] ];
		for (var j = 0; j < y.length; j++)
			sol.push([y[j]]);
		
		var n = Math.floor(tmax / dt);
		for (var i = 1; i <= n; i++) {
			var k1 = deriv(t, y);
			var k2 = deriv(t + dt_2, y.add(k1.mul(dt_2)) );
			var k3 = deriv(t + dt_2, y.add(k2.mul(dt_2)) );
			var k4 = deriv(t + dt, y.add(k3.mul(dt)) );
			t = i*dt;
			
			sol[0].push(t);
			for (var j = 0; j < y.length; j++) {
				y[j] += (dt/6) * (k1[j] + 2 * k2[j] + 2 * k3[j] + k4[j]);
				sol[j+1].push( y[j] );
			}
		}
		return sol;
	}
}

Fourier = {
	square: function(n) {
		if (n === 0)
			return Math.zero;
		
		return function(ωt) { return 4 * Math.sin((2 * n - 1) * ωt) / (Math.PI * (2 * n - 1)); };
	},
	
	triangle: function(n) {
		if (n === 0)
			return Math.zero;
	
		return function(ωt) { return -8 * Math.pow(-1, n) * Math.sin((2 * n - 1) * ωt) / Math.pow(Math.PI * (2 * n - 1), 2); };
	},
	
	sawtooth: function(n) {
		if (n === 0)
			return Math.zero;
		
		return function(ωt) { return 2 * Math.sin(n * ωt) / (Math.PI * n); };
	},
	
	rectified: function(n) {
		if (n === 0)
			return Math.const(2/Math.PI);
		
		return function(ωt) { return -4 * Math.pow(-1, n) * Math.cos(2 * n * ωt) / (Math.PI * (4 * n * n - 1)); };
	},
	
	pulse: function(duty) {
		return function(n) {
			if (n === 0)
				return Math.const(duty);
			
			return function(ωt) { return 2 * Math.sin(n * Math.PI * duty) * Math.cos(n * ωt) / (Math.PI * n); };
		}
	},
	
	series: function(wave, count) {
		var comp = [];
		for (var n = 0; n < count; n++)
			comp.push(wave(n));

		return function(ωt) {
			var res = 0;
			for (var n = 0; n < count; n++)
				res += comp[n](ωt);
			return res;
		}
	}
}

/* OTHER FUNCTIONS */
var isStr = String.isString;
var isNum = Number.isNumber;
var isArr = Array.isArray;
var isFun = Function.isFunction;
var zip = Array.zip;
var time = window.performance && performance.now ? function() { return performance.now(); } : function() { return Date.now(); };


var print = jsCalc.print;
print.digits = 6;
print.term = "\n"; //each print() call ends with this
print.lines = 4; //the CSS height of output window in 'em' units (-1)
print.fix = false; //fixed-point format?


function plot() {
	var a = arguments;
	if (a.length == 0)
		throw "plot() requires at least one argument";
	
	var data = [];
	var opts;
	if ( a.length >= 2 && Array.isArray(a[0]) && Array.isArray(a[1]) ) {
		opts = a[a.length - 1];
		if ( Array.isArray(opts) )
			opts = { };

		var i = 0;
		while (i + 1 < a.length && Array.isArray(a[i]) && Array.isArray(a[i+1])) {
			data.push( { data: Array.zip(a[i], a[i+1]) } );
			i += 2;
		}
	} else if (Function.isFunction(a[0])) {
		opts = a[a.length - 1];
		if (Function.isFunction(a[a.length - 1]) )
			opts = { };

		var x = Array.range( opts.range || plot.range );
		var i = 0;
		while (i < a.length && Function.isFunction(a[i])) {
			data.push( { data: Array.zip(x, x.map( a[i] )) } );
			i++;
		}
	} else if (Array.isArray(a[0]) && a[0].every(Function.isFunction)) {
		opts = a.length > 1 ? a[1] : { };
		var x = Array.range( opts.range || plot.range );
		a[0].forEach(function(f){
			data.push( { data: Array.zip(x, x.map(f)) } );
		});
	} else
		throw "plot: invalid combination of data";
	
	var colors = Array.cycler(opts.colors || plot.colors);
	for (var i = 0; i < data.length; i++)
		data[i].color = colors.next();
	
	var styles = Array.cycler(opts.styles || plot.styles || [], '-');
	for (var i = 0; i < data.length; i++) {
		var s = styles.next();
		if (s.includes('o'))
			data[i].points = { show: true };
		if (s.includes('-'))
		 	data[i].lines = { show: true };
	}
	
	var labels = opts.labels || plot.labels;
	if (String.isString(labels))
		data[0].label = labels;
	else if (Array.isArray(labels))
	{
		var i = 0;
		while (i < data.length && i < labels.length) {
			data[i].label = labels[i];
			i++;
		}			
	}
	
	var shadow = 'shadow' in opts ? opts.shadow : ('shadow' in plot ? plot.shadow : 0);
	for (var i = 0; i < data.length; i++)
		data[i].shadowSize = shadow;
	
	var w0 = $(window).width();
	var h0 = Math.min( $(window).height(), w0 );
	var width = (opts.width || plot.width || Math.floor(0.9 * w0)).toString();
	var height = (opts.height || plot.height || Math.floor(0.9 * h0)).toString();
	var xaxis = { };
	if ("xmin" in opts)
		xaxis.min = opts.xmin;
	if ("xmax" in opts)
		xaxis.max = opts.xmax;
	var yaxis = { };
	if ("ymin" in opts)
		yaxis.min = opts.ymin;
	if ("ymax" in opts)
		yaxis.max = opts.ymax;
	
	var plotWnd = window.open("plot.html", "_blank", "width=" + width +", height=" + height + ", resizable=1, location=0, toolbar=0, menubar=0, scrollbars=0", false);
	plotWnd.data = data;
	plotWnd.opts = { series: { points: { radius: 5 } }, xaxis: xaxis, yaxis: yaxis }; // { yaxis: { autoscaleMargin: 0.01 } };
}

plot.range = [0, 10, 0.1], //default range
plot.colors = ["red", "blue", "#00CC00", "fuchsia", "#00F2FF", "maroon", "teal", "black"]; //default color cycle



