﻿"use strict";

var jsUnits = (function() {
	function Unit(name, fromBase, toBase) {
		this.name = Array.toArray(name);
		this.fromBase = fromBase;
		this.toBase = toBase;
		return this;
	}
	
	Unit.prototype.toString = function() {
		return this.name[0];
	};

	function Categ(name, units) {
		this.name = name;
		this.units = units;
		return this;
	}

	Categ.prototype.findUnit = function(name) {
		for (var k = 0; k < this.units.length; k++)
			if (this.units[k].name.includes(name))
				return this.units[k];

		return null;
	};
		
	Categ.prototype.toString = function() {
		return this.name;
	};

    var all = (function() {
    	function prop(name, factor) { //"normal" proportional unit
			if (!factor)
				factor = 1;
	
			return new Unit( name,
				function(x) { return x / factor; },
				function(x) { return factor * x; }
			);
		}
		
		function inv(name, factor) { //inversely proportional
			return new Unit(name,
				function(x) { return factor / x; },
				function(x) { return factor / x; }
			);
		}
	
		function lin(name, factor, shift) { //mostly for temperature scales
			return new Unit(name,
				function(x) { return (x - shift) / factor; },
				function(x) { return factor * x + shift; }
			);
		}    
    
    	var list = [
	    	new Categ("length", [ prop(['meter','m']), prop('km', 1e3), prop('dm', 0.1), prop('cm', 0.01), prop('mm', 1e-3), prop(['µm','micron'], 1e-6), prop('nm', 1e-9), prop('mile', 1609.344), prop('nautical mile', 1852), prop(['yd','yard'], 0.9144), prop(['ly','light year'], 9.4605284e15), prop(['pc','parsec'], 3.08567758149137e16), prop(['AU','au'], 149597870700), prop(['foot','ft'], 0.3048), prop(['inch','in'], 0.0254), prop(['ångström','angstrom','Å'], 1e-10) ]),
			new Categ("time", [ prop(['second','s']), prop(['min','minute'], 60), prop(['h','hour'], 3600), prop(['d','day'], 86400), prop(['y','year','julian year'], 31557600) ]),
			new Categ("weight", [ prop('kg'), prop('g', 1e-3), prop('mg', 1e-6), prop('µg', 1e-9), prop(['pound','lb'], 0.45359237), prop(['ounce','oz'], 0.028349523125), prop(['metric ton','t'], 1e3), prop(['short ton','ton'], 907.18474), prop(['amu','Da'], 1.660539040e-27), prop('solar mass', 1.98855e30), prop('earth mass', 5.9722e24) ]),
			new Categ("force", [ prop(['newton','N']), prop(['kN','kilonewton'], 1e3), prop(['MN','meganewton'], 1e6), prop('lbf', 4.448222), prop(['dyn','dyne'], 1e-5), prop('kgf', 9.80665) ]),
			new Categ("energy", [ prop(['joule','J']), prop(['kJ','kilojoule'], 1e3), prop(['MJ','megajoule'], 1e6), prop(['mJ','millijoule'], 1e-3), prop(['µJ','microjoule'], 1e-6), prop('erg', 1e-7), prop('cal', 4.184), prop('kcal', 4184), prop('eV', 1.60217662e-19), prop(['BTU','Btu'], 1055.05585), prop('kWh', 3.6e6), prop('MWh', 3.6e9) ]),
			new Categ("power", [ prop(['watt','W']), prop('kW', 1e3), prop('mW', 1e-3), prop(['µW','microwatt'], 1e-6), prop(['MW','megawatt'], 1e6), prop(['hp','horsepower'], 745.699872) ]),
			new Categ("area", [ prop(['m<sup>2</sup>','m²','m^2']), prop('acre', 4046.8564224), prop(['cm<sup>2</sup>','cm²','cm^2'], 1e-4), prop(['mm<sup>2</sup>','mm²','mm^2'], 1e-6), prop(['hectare','ha'], 1e4), prop('are', 100), prop('barn', 1e-28), prop(['ft<sup>2</sup>','ft²','square foot','sq ft','ft^2'], 0.09290304), prop(['in<sup>2</sup>','in²','square inch','sq in','in^2'], 0.00064516), prop(['yd<sup>2</sup>','yd²','sq yd','yd^2'], 0.83612736) ]),
			new Categ("volume", [ prop(['m<sup>3</sup>','m³','m^3']), prop(['cm<sup>3</sup>','cm³','cm^3'], 1e-6), prop(['mm³','mm^3'], 1e-9), prop(['L','liter'], 1e-3), prop(['in<sup>3</sup>','in³','cubic inch','cu in','in^3'], 1.6387064e-5), prop(['ft<sup>3</sup>','ft³','cubic foot','cu ft','ft^3'], 0.028316846592), prop(['UK gallon','imperial gallon'], 0.00454609), prop(['UK pint','imperial pint'], 0.00056826125), prop('US gallon', 0.003785411784), prop('US pint', 0.000473176473), prop(['oil barrel','barrel','bbl'], 0.158987294928) ]),
			new Categ("pressure", [ prop('pascal','Pa'), prop('kPa', 1e3), prop(['mmHg','mm Hg','torr'], 133.322387415), prop('psi', 6894.757), prop(['at','atm'], 101325), prop('bar', 1e5), prop('mbar', 100), prop(['kgf/cm<sup>2</sup>','kgf/cm²','kgf/cm2','kgf/cm^2'], 98066.5) ]),
			new Categ("speed", [ prop('m/s'), prop(['km/h','kph'], 0.277777777777777778), prop('km/s', 1e3), prop('cm/s', 0.01), prop('mph', 0.44704), prop(['knot','kn'], 0.514444444), prop(['speed of light','c'], 299792458) ]),
			new Categ("angle", [ prop('rad'), prop('deg', 0.0174532925199432958), prop('cycle', 6.28318530717958648) ]),
			new Categ("frequency", [ prop(['hertz','Hz']), prop(['kHz'], 1e3), prop(['MHz'], 1e6), prop(['rad/s'], 0.159154943091895336), inv(['second','s'], 1), inv('ms', 1e3) ]),
			new Categ("temperature", [ prop(['kelvin','K']), lin(['°C','celsius'], 1, 273.15), lin(['°F','fahrenheit'], 0.555555555555555556, 255.372222222222222) ]),
			new Categ("wavelength", [ prop('nm'), prop('cm', 1e7), prop('mm', 1e6), prop(['µm','micron'], 1e3), prop('m', 1e9), inv('eV', 1239.841974), inv(['cm<sup>-1</sup>','cm⁻¹','cm^-1','cm-1','wavenumber'], 1e7), inv('Hz', 2.99792458e17), inv('K', 2897772.9) ])
		];
		
		list.findUnit = function(unit) {
			var match;
			if (String.isString(unit))
				match = function(u) { return u.name.includes(unit) };
			else
				match = function(u) { return u === unit; };
		
			for (var i = 0; i < this.length; i++) {
				var list = this[i].units;
				for (var k = 0; k < list.length; k++)
					if ( match(list[k]) )
						return { unit: list[k], cat: this[i] };
			}
	
			return null;
		}
		
		return list;
	})();
	
	function Quantity(value, unit) {
		if ( !(this instanceof Quantity) ) //allow calling 'Quantity()' without 'new'
			return new Quantity(value, unit);
	
		var res = all.findUnit(unit);
		if (!res)
			return "Quantity: unit '" + unit + "' is unknown";
	
		this.value = value;
		this.cat = res.cat;
		this.unit = res.unit;
		return this;		
	}

	Quantity.prototype.toStr = function(fn) {
		return fn(this.value) + " " + this.unit.name[0];
	};
		
	Quantity.prototype.in = function(unit) {
		if (!(unit instanceof Unit)) {
			var name = unit;
			var unit = this.cat.findUnit(name);
			if (!unit)
				return "Quantity.in: unit '" + name + "' is unknown";
		}
	
		return unit.fromBase(this.unit.toBase(this.value));
	};
	
	Quantity.prototype.to = function(unit) {
		if (!(unit instanceof Unit)) {
			var name = unit;
			var unit = this.cat.findUnit(name);
			if (!unit)
				return "Quantity.to: unit '" + name + "' is unknown";
		}
	
		return new Quantity(unit.fromBase(this.unit.toBase(this.value)), unit);
	};
	
	Quantity.prototype.add = function() {
		var a = arguments;
		var q;
		if (a.length == 1 && (a[0] instanceof Quantity))
			q = a[0];
		else if (a.length == 2)
			q = new Quantity(a[0], a[1]);
		else
			throw "Quantity.add: invalid parameters";
		
		if (q.cat !== this.cat)
			throw "Quantity.add: dimension mismatch";
		
		var base = this.cat.units[0];	
		var value = this.in(base) + q.in(base);
		return new Quantity(value, base);
	};
	
	Quantity.prototype.sub = function() {
		var a = arguments;
		var q;
		if (a.length == 1 && (a[0] instanceof Quantity))
			q = a[0];
		else if (a.length == 2)
			q = new Quantity(a[0], a[1]);
		else
			throw "Quantity.add: invalid parameters";
		
		if (q.cat !== this.cat)
			throw "Quantity.add: dimension mismatch";
		
		var base = this.cat.units[0];	
		var value = this.to(base) - q.to(base);
		return new Quantity(value, base);
	};
		
	var regex = (function() {
		var value = "([+-]?[0-9.][^\\s]*|'[^']+')";
		var space = "\\s+";
		var unit = "([^\\s']+|'[^']+')";
		var arrow = "[\\-]*\\>";
		var digits = "(?:\\s+[dD#]([0-9]+))?";
		
		return new RegExp("^" + value + space + unit + space + arrow + space + unit + digits);
	})();
	
	function exec(input, out) {
		var parts = regex.exec(input);
		if (parts == null) {
			append("error: invalid syntax");
			return;
		}
		
		var fromVal;
		if (parts[1].charAt(0) == "'") {
			fromVal = eval(parts[1].strip("'"));
			if (!Number.isNumber(fromVal) || Number.isNaN(fromVal)) {
				append("error: could not evaluate " + parts[1]);
				return;
			}	
		} else {
			fromVal = Number(parts[1]);
			if (Number.isNaN(fromVal)) {
				append("error: not a number (" + parts[1] + ")");
				return;
			}
		}
		
		var fromUnit = parts[2].strip("'");
		var toUnit = parts[3].strip("'");
		var matches = 0;
		
		var digits = 6;
		if (parts[4]) {
			var digits = Number(parts[4]);
			if (digits < 2)
				digits = 2;
			else if (digits > 15)
				digits = 15;
		};
		
		out.empty();
		for (var i = 0; i < all.length; i++) {
			var u = all[i].units;
			for (var j = 0; j < u.length; j++) {
				if (u[j].name.includes(fromUnit)) {
					var std = u[j].toBase(fromVal);
					for (var k = 0; k < u.length; k++) {
						if (u[k].name.includes(toUnit)) {
							var toVal = u[k].fromBase(std).toPre(digits);
							out.append(fromVal.toPre(digits) + " " + u[j].name[0] + " corresponds to " + toVal + " " + u[k].name[0] + " (category: " + all[i].name + ")<br/>");
							matches++;
							break;
						}
					}
				}
			}
		}
		if (matches == 0)
			out.append("units '" + fromUnit + "' and/or '" + toUnit + "' are unknown or relate to different dimensions<br/>");
	}
	
	function convert(value, fromName, toName) {
		var res = all.findUnit(fromName);
		if (!res)
			return "unit '" + fromName + "' is unknown";
		
		var fromUnit = res.unit;
		var toUnit = res.cat.findUnit(toName);
		if (!toUnit)
			return "unit '" + toName + "' is unknown";
		
		return toUnit.fromBase(fromUnit.toBase(value));
	}
	
	return { convert: convert, exec: exec, Unit: Unit, Quantity: Quantity };
})();