﻿"use strict";

if (!Object.create) {
	Object.create = function (proto, properties) {
		if (!(proto === null || typeof proto === "object" || typeof proto === "function"))
			throw new TypeError('Object.create: argument must be an object or null');

		var obj = new Object();
		obj.__proto__ = proto;
		if (typeof properties === "object")
			Object.defineProperties(obj, properties);

		return obj;
	};
}

Object.prototype.prec = function(digits) {
	if (Function.isFunction(this.toStr))
		return this.toStr(Number.floatFormat(digits));
	
	return this.toString();
};

Object.prototype.fixed = function(digits) {
	if (Function.isFunction(this.toStr))
		return this.toStr(Number.fixedFormat(digits));
	
	return this.toString();
};

Number.isNumber = function(obj) {
	return (typeof obj === 'number') || (typeof obj === 'object' && obj !== null && obj.constructor === Number);
}

if (!Number.isFinite)
	Number.isFinite = function(obj) { return Number.isNumber(obj) && isFinite(obj); };

if (!Number.isNaN)
	Number.isNaN = function(obj) { return obj !== obj; };

if (!Number.isInteger)
	Number.isInteger = function(obj) { return Number.isFinite(obj) && Math.floor(obj) === obj; };

Number.prototype.pow = function(x) {
	return Math.pow(this, x);
}

Number.prototype.root = function(x) {
	return Math.pow(this, 1/x);
}

Number.prototype.toStr = function(fn) {
	return fn(this);
}

Number.objToStr = function(obj, fn) {
	if (obj === null)
		return 'null';
	
	if (typeof obj === 'undefined')
		return 'undefined';
	
	if (Function.isFunction(obj.toStr))
		return obj.toStr(fn);
	
	return obj.toString();
}

Number.floatFormat = function(digits) {
	if (!Number.isFinite(digits))
		digits = 15; //set default value of digits here
	else if (digits < 2)
		digits = 2;
	else if (digits > 15)
		digits = 15;

	return function(obj) {
		if (!Number.isNumber(obj))
			throw "Number.floatFormat: number expected";
		
		if (typeof obj === 'object')
			obj = obj.valueOf();
		
		if (obj === 0)
			return "0";
		
		if (!Number.isFinite(obj))
			return obj.toString();
			
		var value = ( digits === 15 ? obj : parseFloat(obj.toPrecision(digits)) );
		var abs = Math.abs(value);
		if (abs < 1e-3)
			return value.toExponential();
		
		if (abs < 1e7 || Math.floor(Math.log10(abs)) < digits)
			return value.toString();
		
		return value.toExponential().replace("e+", 'e');
	};
}

Number.fixedFormat = function(digits) {
	if (!Number.isFinite(digits))
		digits = 0; //set default value of digits here
	else if (digits > 15)
		digits = 15;

	return function(obj) {
		if (!Number.isNumber(obj))
			throw "Number.fixedFormat: number expected";
	
		return obj.toFixed(digits);
	}
}

Function.isFunction = function(obj) {
	return (obj instanceof Function) || (typeof obj === 'function');
}

String.isString = function(obj) {
	return (typeof obj === 'string') || (typeof obj === "object" && obj !== null && obj.constructor === String);
}

if (!String.prototype.repeat)
	String.prototype.repeat = function(n) {
		if (!Number.isFinite(n))
			throw new TypeError("repeat count must be a number");
		
		if (n < 0)
			throw new RangeError("repeat count must be non-negative");
		
		var res = "";
		for (var i = 1; i <= n; i++)
			res += this;
		return res;
	}

if (!String.prototype.trim)
	String.prototype.trim = function() { return this.replace(/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g, ''); };

if(!String.prototype.includes)
	String.prototype.includes = function(str, start) { return this.indexOf(str, start) !== -1; };

if (!String.prototype.startsWith)
	String.prototype.startsWith = function(str){
		return this.substr(0, str.length) === str;
};

if (!String.prototype.endsWith)
	String.prototype.endsWith = function(str) {
		return this.substr(this.length - str.length, str.length) === str;
	};

String.prototype.strip = function(s) {
	if (this.startsWith(s) && this.endsWith(s)) {
		var len = s.length;
		return this.slice(len, -len);
	}
	
	return this;
};

String.prototype.pad = function(len, c, alignLeft) {
	if (!c)
		c = ' ';
	else if (!String.isString(c) || c.length !== 1)
		throw "String.lpad: argument 'c' is invalid (single character expected)";
	
	if (String.isString(len))
		len = len.length;
	else if (!Number.isNumber(len))
		throw "String.lpad: argument 'len' is invalid (number or string expected)";
	
	var count = len - this.length;
	return count <= 0 ? this : (alignLeft ? this + c.repeat(count) : c.repeat(count) + this);
}

if (!Math.log10)
	Math.log10 = function(x) { return Math.log(x) / Math.LN10; };

if (!Math.sign)
	Math.sign = function(x) {
		return ((x > 0) - (x < 0)) || +x;
		//return typeof x === 'number' ? x ? x < 0 ? -1 : 1 : x === x ? 0 : NaN : NaN;
	};

Math.sum = function() {
	var count = 0;
	var Σ = 0;
	
	function parse(array) {
		for (var i = 0; i < array.length; i++) {
			var val = array[i];
			if (Number.isNumber(val) && !Number.isNaN(val)) {
				count++;
				Σ += val;
			}
			else if (Array.isArray(val))
				parse(val);
		}
	}

	parse(arguments);
	if (count == 0)
		throw "Math.sum: at least one numerical argument required";
	
	return Σ;
};

Math.mean = function() {
	var count = 0;
	var Σ = 0;
	
	function parse(array) {
		for (var i = 0; i < array.length; i++) {
			var val = array[i];
			if (Number.isNumber(val) && !Number.isNaN(val)) {
				count++;
				Σ += val;
			}
			else if (Array.isArray(val))
				parse(val);
		}
	}

	parse(arguments);
	if (count == 0)
		throw "Math.mean: at least one numerical argument required";
	
	return Σ / count;
};

Math.stdev = function() {
	var count = 0;
	var Σx = 0;
	var Σxx = 0;
	
	function parse(array) {
		for (var i = 0; i < array.length; i++) {
			var val = array[i];
			if (Number.isNumber(val) && !Number.isNaN(val)) {
				count++;
				Σx += val;
				Σxx += val * val;
			}
			else if (Array.isArray(val))
				parse(val);
		}
	}

	parse(arguments);
	if (count == 0)
		throw "Math.stdev: at least one numerical argument required";
	
	return Math.sqrt( Σxx / count - Σx * Σx / count / count );
};

Math.zero = function(x) { return 0; };

Math.const = function(c) { return function(x) { return c; }; };

Math.perform = function(a, b, op) {
	if (Number.isNumber(a) && Number.isNumber(b))
		return op(a,b);
	
	if (Array.isArray(a) && Number.isNumber(b))
		return a.map(function(x) { return Math.perform(x, b, op); });

	if (Number.isNumber(a) && Array.isArray(b))
		return b.map(function(x) { return Math.perform(a, x, op); });

	if (Array.isArray(a) && Array.isArray(b)) {
		var res = [];
		var m = Math.min(a.length, b.length);
		for (var i = 0; i < m; i++)
			res.push( Math.perform(a[i], b[i], op) );
		
		return res;
	}
		
	throw "Math.perform: cannot perform operation with these data types";
};

Math.binops = {
	add: function(x,y){ return x + y; },
	sub: function(x,y){ return x - y; },
	mul: function(x,y){ return x * y; },
	div: function(x,y){ return x / y; }
}

if (!Array.isArray)
	Array.isArray = function(obj) { return Object.prototype.toString.call(obj) === '[object Array]'; };

Array.toArray = function(obj) { return Array.isArray(obj) ? obj : [obj]; };

Array.linspace = function(start, stop, count) {
	if (typeof start === 'object')
		start = start.valueOf();
	
	if (typeof stop === 'object')
		stop = stop.valueOf();
	
	if (typeof count === 'object')
		count = count.valueOf();
	
	if ( !Number.isFinite(start) || !Number.isFinite(stop) || !Number.isFinite(count) )
		throw "Array.linspace: invalid arguments (numbers expected)";
	
	if (!Number.isInteger(count) || count < 2)
		throw "Array.linspace: count must be integer >= 2";
	
	var step = (stop - start) / (count - 1);
	var res = [];
	for (var i = 0; i < count; i++)
		res.push(start + i * step);
	return res;
}

Array.const = function(len, val) {
	if (typeof len === 'object')
		len = len.valueOf();
	
	if (!Number.isFinite(len) || len < 0 || !Number.isInteger(len))
		throw "Array.const: invalid array length (must be integer >= 0)";
	
	if (typeof val === 'undefined')
		val = 0;

	var res = new Array(len);
	for (var i = 0; i < len; i++)
		res[i] = val;
	return res;
	/* alternatively: Array.apply(null, Array(len) ).map(Number.prototype.valueOf, val) */
}

Array.range = function() {
	var a = arguments;
	if ( a.length === 0 )
		throw "Array.range: at least 1 argument expected";
		
	if ( a.length === 1 && Array.isArray(a[0]))
		a = a[0];
	
	var start;
	var stop;
	var step;
	if ( a.length === 1 && Number.isFinite(a[0]) ) {
		start = 0;
		stop = a[0];
		step = 1;
	}
	else if ( a.length === 2 && Number.isFinite(a[0]) && Number.isFinite(a[1]) ) {
		start = a[0];
		stop = a[1];
		step = 1;
	}
	else if ( a.length === 3 && Number.isFinite(a[0]) && Number.isFinite(a[1]) && Number.isFinite(a[2]) ) {
		start = a[0];
		stop = a[1];
		step = a[2];
	}
	else
		throw "Array.range: invalid arguments";

	var count = (stop - start) / step;
	if (!Number.isFinite(count) || count < 0)
			throw "Array.range: division by zero or step sign error";

	var res = [];
	for (var i = 0; i <= count; i++)
		res.push(start + i*step);
	
	return res;
}

Array.zip = function() {
	var m = Number.POSITIVE_INFINITY;
	for (var i = 0; i < arguments.length; i++) {
		if ( !Array.isArray(arguments[i]) )
			throw "Array.zip: invalid arguments (arrays expected)";
		if (arguments[i].length < m)
			m = arguments[i].length;
	}

	var res = [];
	for (var k = 0; k < m; k++) {
		var row = [];
		for (var i = 0; i < arguments.length; i++)
			row.push(arguments[i][k]);
		res.push(row);
	}
	
	return res;
}

Array.cycler = function(items, def) {
	if (!Array.isArray(items))
		items = [items];
	else if (items.length === 0 && (typeof def === 'undefined'))
		throw "Array.cycler: the number of elements in 'items' is zero";
	
	var index = 0;
	
	if (typeof def === 'undefined')
		return {
			next: function() { //infinitely cycle through the array if no default value given
				var res = items[index];
				index = (index + 1) % items.length;
				return res;
			},
			reset: function() { index = 0; }
		};
	
	return {
		next: function() { return index < items.length ? items[index++] : def; },
		reset: function() { index = 0; }
	};
}

Array.prototype.nonZero = function(fn) {
		for (var n = 0; n < this.length; n++)
			if ( this[n] )
				fn(this[n], n);
	};

Array.prototype.toStr = function(fn) {
	return "[" + this.map(function(x) { return Number.objToStr(x, fn); }).join(", ") + "]";
}

Array.prototype.clone = function() {
	return this.map(function(x) {
		return Array.isArray(x) ? x.clone() : x;
	});
}

Array.prototype.extend = function (array) {
    if (!Array.isArray(array))
    	throw "Array.extend: array expected";
    
    array.forEach(function(x) { this.push(x); }, this);
    return this;   
}

Array.prototype.equals = function(x){
	if (!Array.isArray(x) || this.length !== x.length)
		return false;
	
	for (var i = 0; i < this.length; i++) {
		if (Array.isArray(this[i]) && Array.isArray(x[i])){
			if (!this[i].contentEquals(x[i]))
				return false;       
		} else if (this[i] !== x[i])
		    return false;   
	}

    return true;
}

if (!Array.prototype.find)
	Array.prototype.find = function(predicate, context) {
		if (!Function.isFunction(predicate))
			throw new TypeError('predicate must be a function');
		
		var len = this.length;
		for (var i = 0; i < len; i++)
			if (predicate.call(context, this[i], i, this))
				return this[i];
		
		return undefined;
	}


if (!Array.prototype.fill)
	Array.prototype.fill = function(val, start, end) {
		if (typeof start === 'undefined')
			start = 0;
		else if (!Number.isInteger(start))
			throw new TypeError("start index must be an integer");
		else if (start < 0)
			start += this.length;
		
		if (typeof end === 'undefined')
			end = this.length;
		else if (!Number.isInteger(end))
			throw new TypeError("end index must be an integer");
		else if (end < 0)
			end += this.length;
		
		for (var i = start; i < end; i++)
			this[i] = val;
		return this;
	}

Array.prototype.add = function(a) {
	return Math.perform(this, a, Math.binops.add);
}

Array.prototype.sub = function(a) {
	return Math.perform(this, a, Math.binops.sub);
}

Array.prototype.mul = function(a) {
	return Math.perform(this, a, Math.binops.mul);
}

Array.prototype.div = function(a) {
	return Math.perform(this, a, Math.binops.div);
}

var Cookie = {
	write: function(name, value, days) {
				var expires;
				if (days) {
					var date = new Date();
					date.setTime(date.getTime() + days * 86400000);
					expires = "; expires=" + date.toGMTString();
				}
				else 
					expires = "";
				document.cookie = name + "=" + encodeURIComponent(value) + expires + "; path=/";
			},

	read: 	function(name) {
				var res = ( new RegExp('[; ]' + name + '=([^\\s;]*)') ).exec( ' ' + document.cookie );
				if (res)
					return decodeURIComponent(res[1]);
				
				return null;
			},
				
	erase:	function(name) {
				this.write(name, "", -1);
			}
}