﻿function Polynom(coeff) {
	if (!Array.isArray(coeff) || coeff.length === 0)
		throw "Polynom: invalid coefficients (non-empty array expected)";

	function pol(x) {
		var res = 0;
		for (var n = 0; n < coeff.length; n++)
			res += coeff[n] * Math.pow(x, n);
		return res;
	}
	
	pol.toStr = function(fn) { return Polynom._format(coeff, fn); };
	
	return pol;
}

Polynom.eval = function(coeff, x) {
	if (!Array.isArray(coeff) || coeff.length === 0)
		throw "Polynom: invalid coefficients (non-empty array expected)";

	if (!Number.isNumber(x))
		throw "Polynom: x must be a number";
	
	var res = 0;
	for (var n = 0; n < coeff.length; n++)
		res += coeff[n] * Math.pow(x, n);
	return res;
}

Polynom.base = function(n) {
		if (typeof n === 'object')
			n = n.valueOf();
	
		if (!Number.isInteger(n))
			throw "Polynom.base: invalid argument (integer expected)";
	
		if (n == 0)
			return Math.const(1);
		
		if (n == 1)
			return function(x) { return x; }
		
		return function(x) { return Math.pow(x, n); }
	};

Polynom.baseSet = function(degree) {
		if (typeof degree === 'object')
			degree = degree.valueOf();
	
		if (!Number.isInteger(degree) || degree < 0)
			throw "Polynom.baseSet: invalid degree";

		var res = [];
		for (var n = 0; n <= degree; n++)
			res.push( Polynom.base(n) );
			
		return res;
	};

/*Polynom.powersv = function(degree) {
		if (typeof degree === 'object')
			degree = degree.valueOf();
	
		if (!Number.isInteger(degree) || degree < 0)
			throw "Polynom.vectorOfPowers: invalid degree";

		return function(x) {
			var res = [];
			for (var n = 0; n <= degree; n++)
				res.push( Math.pow(x, n) );
				
			return res;
		}
	};*/
	
Polynom._format = function(coeff, fn) {
		var res = [];
		coeff.nonZero(function(c, n) {
			var a = fn(c);
			var p = n === 0 ? '' : (n < 2 ? 'x' : 'x^' + n);
			res.push( (c < 0 ? '(' + a + ')' : a) + p);
		});
		
		return 'y = ' + res.join(' + ');
	};

function Cheb(coeff) {
	if (!Array.isArray(coeff) || coeff.length === 0)
		throw "Cheb: invalid coefficients (non-empty array expected)";

	if (coeff.length === 1)
		return function(x) { return coeff[0]; };
	
	if (coeff.length === 2)
		return function(x) { return coeff[0] + coeff[1] * x; };
	
	function pol(x) {
		var res = coeff[0] + coeff[1] * x;
		var T_1 = x;
		var T_2 = 1;
		
		for (var n = 2; n < coeff.length; n++) {
			var T = 2 * x * T_1 - T_2;
			res += coeff[n] * T;
			T_2 = T_1;
			T_1 = T;
		}
		
		return res;
	}
	
	pol.toStr = function(fn) { return Cheb._format(coeff, fn); };
	
	return pol;
}

Cheb.base = function(n) {
		if (typeof n === 'object')
			n = n.valueOf();
	
		if (!Number.isInteger(n))
			throw "Cheb.base: invalid argument (integer expected)";
	
		if (n == 0)
			return Math.const(1);
		
		if (n == 1)
			return function(x) { return x; }
	
		return function(x) {
			var res;
			var T_1 = x;
			var T_2 = 1;
			
			for (var i = 2; i <= n; i++) {
				res = 2 * x * T_1 - T_2;
				T_2 = T_1;
				T_1 = res;
			}
			
			return res;
		};
	}

Cheb.baseSet = function(degree) {
		if (typeof degree === 'object')
			degree = degree.valueOf();
	
		if (!Number.isInteger(degree) || degree < 0)
			throw "Cheb.baseSet: invalid degree";

		var res = [];
		for (var n = 0; n <= degree; n++)
			res.push( Cheb.base(n) );
			
		return res;
	};

Cheb._format = function(coeff, fn) {
		if (!Function.isFunction(fn))
			throw "Cheb._format: number format function not provided";
		
		var res = [];
		coeff.nonZero(function(c, n) {
			var a = fn(c);
			res.push( (c < 0 ? '(' + a + ')' :  a ) + '*T_' + n + '(x)' );
		});
		
		return 'y = ' + res.join(' + ');
	};
